/*ILDAPARSER.C

  CMPE 490 w2010 - Group 9
  Written by Michael Ivey

  PARAMETERS: ./ildaparser FILE1 FILE2 FILE3 FILE4

  This program parses the International Laser Display Association's
  (ILDA) file format *.ild - this is a binary file format for which the
  specs can be found here:
  http://www.laserist.org/StandardsDocs/IDTF05-finaldraft.pdf

  The program also supports reading into memory a particular custom type
  of X by X raster image, *.ras, which has a format of the following:

  #rows/cols #frames
  0 0 0 0 0 0 <-------Each zero represents the laser off
  0 1 0 0 0 0 <-------Files must be an n x n matrix of 0s and 1s

  Files can be exported in differing ways depending on which functions
  are enabled-  they can be exported as binary objects,  binary arrays
  or a text file containing variable declarations which can be 
  copied/pasted into a C file.

  Note that ILDA files are also downsampled to fit on our particular laser
  projector - while .ild files have a full -32867-32867 range, our laser
  projector is only capable of 0-1023 and as such all values must be adjusted
  and downsampled.

  All output is exported to the file "images.let".
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#define TRUE 1
#define FALSE 0
#define MAXDIM 1024;

/*Define structures for organising data*/

/*Represents a single point in an ILDA image*/
 typedef struct point
{
	short int x;
	short int y;
	int isOn;
} point;

/*Represents a frame in an ILDA image*/
 typedef struct frame
{
	int point_count;
	point* points;
	int is_3D;
} frame;

/*Represents an image comprised of multiple frames*/
typedef struct image
{
	frame* frames;
	int frame_count;
	char* author;
	char* name;
}image;

/*Parses a raster file as defined above and populates an image struct with the file data*/
int parseRaster( char* filename, image* newimage)
{
	FILE* raster_file;
	
	raster_file = fopen(filename,"r");
	int framecount;
	int gridsize;
	image picture;
	/*First row has the number of cols/rows and # of frames */
	if(fscanf(raster_file,"%d",&gridsize) == EOF) return -1;
	if(fscanf(raster_file,"%d",&framecount) == EOF) return -1;

	int interval = MAXDIM;
	interval = interval/(gridsize-1);
	picture.frames = malloc(framecount*sizeof(frame));
	picture.frame_count = framecount;
	picture.name = filename;

	/*Loop every grid*/
	int frame;
	int point = FALSE;
	for(frame=0;frame<framecount;frame++)
	{
		picture.frames[frame].points=malloc(2*gridsize*gridsize*sizeof(struct point));
		picture.frames[frame].is_3D = FALSE;
		picture.frames[frame].point_count = 2*gridsize*gridsize;
		/*Do this for every grid*/
		int counter = 0;
		int i;
		for(i=0;i<gridsize;i++)
		{
			int j;
			for(j=0;j<gridsize;j++)
			{
				if(fscanf(raster_file,"%d",&point) ==EOF) return -1;
				/*Store a new point*/
				picture.frames[frame].points[counter].x = interval*j ;
				picture.frames[frame].points[counter].y = interval*i ;
				if(point==TRUE)
					picture.frames[frame].points[counter].isOn = TRUE;
				else picture.frames[frame].points[counter].isOn = FALSE;
				counter++;
				/*INSERT BLANK POINT*/
				picture.frames[frame].points[counter].x = interval*j ;
				picture.frames[frame].points[counter].y = interval*i ;
				picture.frames[frame].points[counter].isOn = FALSE;
				counter++;
			}
		}
	}
	fclose(raster_file);
	*newimage = picture;
	return 0;
}

/*Parses an .ild file into the given image object*/
int parseILDA( char* filename, image* newimage )
{
	FILE* ilda_file;
	unsigned char intbuffer[2]="";
	char buffer[4]="";
	char skipbuffer[16]="";

	image picture;
	picture.frame_count = 1;
	picture.name = filename;
	frame* framelist;
	ilda_file = fopen (filename,"r");
	if (ilda_file == NULL) return 1;

	int fread_status;

	int fr;
	int count = 1;/*This will be changing once we read it in*/

	for(fr=0;fr<count;fr++)
	{

		int is3D;
		int pointCount;

		/* Check to see if we have a valid frame - if not throw error*/
		fread_status = fread(buffer,sizeof(char),4,ilda_file);
		if (fread_status == EOF) break;
		/*printf("STUFF:'%s'\n",buffer);*/
		if (strncmp(buffer,"ILDA",4) != 0) return 2;

		/* Check to see if it's a 2D or 3D image (different processing) */
		fread(buffer,sizeof(char),4,ilda_file);

		if ((buffer [3]) == 0x00) /*3D*/
			is3D = TRUE;
		else if ((buffer[3] ) == 0x01 ) /*2D*/
			is3D = FALSE;
		else return -1; /* We don't handle this case */

		/*Have we reached EOF?*/
		fread(skipbuffer,sizeof(char),16,ilda_file); /* Skip bits 9-24 */
		printf("'%s'\n",skipbuffer);
		if (strcmp(skipbuffer,"END OF FILE") == TRUE)
		{
			break;
		}

		/* Get point count */
		fread(intbuffer,sizeof(char),2,ilda_file); /* Read bits 25-26 */
		pointCount = (int)(intbuffer[0]<<8)+ (int)intbuffer[1];
		if (pointCount == 0) /* End of ILDA file */
		{
			break;
		}
		point* pointlist = malloc(pointCount*sizeof(point));

		/* Read in points */
		fread(skipbuffer,sizeof(char),2,ilda_file); /* Skip bits 27-28 */

		/*Get total framecount and set this if first frame*/
		fread(intbuffer,sizeof(char),2,ilda_file); /* Read bits 29-30 */
		count = (int)(intbuffer[0]<<8)+ (int)intbuffer[1];

		if(fr==0)/*Only run once*/
		{
			framelist = malloc(count*sizeof(frame));
			picture.frame_count = count;
			printf("SET COUNT TO %d\n",picture.frame_count);
		}
		framelist[fr].point_count = pointCount;
		framelist[fr].is_3D = is3D;


		fread(skipbuffer,sizeof(char),2,ilda_file); /* Skip bits 31-32 */

		int i;
		int tempint;
		for(i=0;i<framelist[fr].point_count;i++)
		{
			fread(intbuffer,sizeof(char),2,ilda_file); /* Read X-bits */
			pointlist[i].x = (int)(intbuffer[0]<<8)+ (int)intbuffer[1];

			fread(intbuffer,sizeof(char),2,ilda_file); /* Read Y-bits */
			pointlist[i].y = (int)(intbuffer[0]<<8)+ (int)intbuffer[1];

			/*Cut out precision; adds enough to positivize*/

			tempint = pointlist[i].x + 32768;
			pointlist[i].x = tempint >> 6;

			tempint = pointlist[i].y + 32768;
			pointlist[i].y = tempint >> 6;

			/*If a 3D image, skip the Z info bits*/
			if (framelist[fr].is_3D == TRUE)
				fread(intbuffer,sizeof(char),2,ilda_file); /* Read Z-bits */

			fread(intbuffer,sizeof(char),2,ilda_file); /* Read status bits */
			if (intbuffer[0] & 0x40) pointlist[i].isOn = FALSE;
			else pointlist[i].isOn = TRUE;
			/*printf("   - %d: %d %d %d\n",i,pointlist[i].x,pointlist[i].y,pointlist[i].isOn);*/

		}
		framelist[fr].points = pointlist;

		/* Move nextframe pointer */
		/*printf("|%d| |%d| |%d|\n",framelist[fr].point_count,framelist[fr].points[0].x,framelist[fr].is_3D);*/
	}
	picture.frames = framelist;
	*newimage = picture;
	fclose (ilda_file);

	return 0;

}

/*Reads a binary list of images. Intended for testing to make sure that a binary file
  containing a list of image structs to be read into memory was written properly, but
  it can also be used as the basis for a parser for these types of exported files.*/
int readBinary(image** imagelist) {
  FILE *inFile;


  int imagecount;
  image* images;

  if(!(inFile = fopen("images.let", "r")))
    return 1;

  fread(&imagecount,sizeof(int),1,inFile);

  printf("READING IMAGECOUNT OF %d\n",imagecount);

  images = malloc(imagecount*sizeof(image));
  fread(images,sizeof(image),imagecount,inFile);
  fclose(inFile);

	/*DEBUG CODE HERE - UNCOMMENT IF YOU WANT TO USE*/
	/*int i;
	for(i=0;i<imagecount;i++)
	{
		int fr;
		printf("FRAMECOUNT = %d\n",images[i].frame_count);
		for(fr=0;fr<images[i].frame_count;fr++)
		{
			printf("----FRAME %d/%d: %d\n",fr+1,images[i].frame_count,
					images[i].frames[fr].point_count);
			int j;
			for(j=0;j<images[i].frames[fr].point_count;j++)
			{
				printf("%d ",images[i].frames[fr].points[j].y);
			}
			printf("\n");
		}

	}*/

	*(imagelist) = images;
	printf("Exported File OK.\n");
	return imagecount;
}

/*Memory clean up - frees all the image objects created to read in a .ild file*/
int freeImageList(image* images, int count)
{
	int i;
	for (i=0;i<count;i++)
	{
		int fr;
		for (fr=0;fr<images[i].frame_count;fr++)
		{
			free(images[i].frames[fr].points);
		}
		free(images[i].frames);
	}
	free(images);
	return 0;
}

/*Reads a file containing a bunch of C arrays in binary format into memory.
  This is used for a doublecheck to ensure files were written properly.
  It can also be used as the basis for writing a proper parser.*/
int readArrays( char* filename )
{

	/*Read in this format:

	imagecount; (int)
	framecount pointcount; (ints)
	xpoints;
	ypoints;
	onoff;

	*/

	FILE* file;
	if(!(file=fopen(filename,"r"))) return -1;

	int i;
	int count;
	fread(&(count),sizeof(int),1,file);
	printf("%d IMAGES EXIST:\n",count);
	for (i=0;i<count;i++)
	{	
		int frames;
		fread(&frames,sizeof(int),1,file);
		int fr;
		for (fr=0;fr<frames;fr++)
		{
			int points;
			fread(&(points),sizeof(int),1,file);

			short int* x = malloc(points*sizeof(short int));
			short int* y = malloc(points*sizeof(short int));
			short int* on = malloc(points*sizeof(short int));

			fread(x,sizeof(short int),points,file);
			fread(y,sizeof(short int),points,file);
			fread(on,sizeof(short int),points,file);
			printf("FRAME %d: %d %d %d - %d points\n",(int)fr,(int)x[0],(int)y[0],(int)on[0],points);
		
			free(x);
			free(y);
			free(on);
		}
	}
	fclose(file);	
	return 0;
}

/*Writes a plaintext file declaring multiple shorts and char arrays that contain]
  all the points for the image.  This is what we currently use for the laser projector*/
int spitOutPlain( image* images, int count, char* filename)
{

	/*Spit out this format:

	imagecount; (int)
	framecount (int)
	pointcount; (ints)
	xpoints;
	ypoints;
	onoff;

	*/
	FILE* file;
	if(!(file=fopen(filename,"w"))) return -1;

	int i;
	int j;
	for (i=0;i<count;i++)
	{	
		int fr;
		fprintf(file,"u_int image%d_frames = %d;\n",i,images[i].frame_count);

		fprintf(file,"u_int image%d_points[] = {",i);
		for (fr=0;fr<images[i].frame_count;fr++)
		{
			int points = images[i].frames[fr].point_count;
			fprintf(file,"%d,",points);


		}
		fprintf(file,"};\n");

		fprintf(file,"u_short image%d_x[][] = {",i);
		for (fr=0;fr<images[i].frame_count;fr++)
		{
			int points = images[i].frames[fr].point_count;
			fprintf(file,"\n{");

			for(j=0;j<points;j++)
			{
				fprintf(file,"%d,",images[i].frames[fr].points[j].x);
			}
			fprintf(file,"},");

		}
		fprintf(file,"};\n");

		fprintf(file,"u_short image%d_y[][] = {",i);
		for (fr=0;fr<images[i].frame_count;fr++)
		{
			int points = images[i].frames[fr].point_count;
			fprintf(file,"\n{");

			for(j=0;j<points;j++)
			{
				fprintf(file,"%d,",images[i].frames[fr].points[j].y);
			}
			fprintf(file,"},");

		}
		fprintf(file,"};\n");

		fprintf(file,"char image%d_on[][] = {",i);
		for (fr=0;fr<images[i].frame_count;fr++)
		{
			int points = images[i].frames[fr].point_count;
			fprintf(file,"\n{");

			for(j=0;j<points;j++)
			{
				fprintf(file,"%d,",images[i].frames[fr].points[j].isOn);
			}
			fprintf(file,"},");

		}
		fprintf(file,"};\n");
	}


	fclose(file);	
	return 0;
}



/*Writes a binary file consisting of various arrays containing all the image point data.*/
int spitOutArrays( image* images, int count, char* filename)
{

	/*Spit out this format:

	imagecount; (int)
	framecount (int)
	pointcount; (ints)
	xpoints;
	ypoints;
	onoff;

	*/
	FILE* file;
	if(!(file=fopen(filename,"w"))) return -1;

	int i;
	fwrite(&(count),sizeof(int),1,file);

	for (i=0;i<count;i++)
	{	
		fwrite(&(images[i].frame_count),sizeof(int),1,file);
		int fr;
		for (fr=0;fr<images[i].frame_count;fr++)
		{
			int points = images[i].frames[fr].point_count;

			fwrite(&(points),sizeof(int),1,file);

			short int* x = malloc(points*sizeof(short int));
			short int* y = malloc(points*sizeof(short int));
			short int* on = malloc(points*sizeof(short int));
			int j;
			for(j=0;j<points;j++)
			{
				x[j] = images[i].frames[fr].points[j].x;
				y[j] = images[i].frames[fr].points[j].y;
				on[j] = images[i].frames[fr].points[j].isOn;
			}
			fwrite(x,sizeof(short int),points,file);
			fwrite(y,sizeof(short int),points,file);
			fwrite(on,sizeof(short int),points,file);

			free(x);
			free(y);
			free(on);
		}
	}
	fclose(file);	
	return 0;
}

/*Writes a binary file containing a list of all the image structs created to be read back in*/
int writeBinary(image* images, int imagecount, char* filename)
{
	/*Writes all the objects to a file*/
	FILE* outfile;

	if(!(outfile = fopen(filename, "w")))
	   return 1;

	/* use fwrite to write binary data to the file */
	fwrite(&(imagecount),sizeof(int),1,outfile);
	fwrite(images,sizeof(image),imagecount,outfile);
	fclose(outfile);

	return 0;
}

/*Main function. Takes the files passed in, parses them based on type and exports them.*/
int main( int argc, char* argv[])
{
	image* images;
	int i;

	int imagecount = argc-1;
	images = malloc(imagecount*sizeof(image));

	for (i=0;i<imagecount;i++)
	{
		images[i].name = argv[i+1];
		int code;
		if (strstr(argv[i+1],".ild"))
			code = parseILDA(argv[i+1],&(images[i]));
		else if (strstr(argv[i+1],".ras"))
			code = parseRaster(argv[i+1],&(images[i]));
		else code = -1;
		if (code != 0)
		{
			printf("ERROR PARSING ON %d; CODE %d\n",i,code);
			return -1;
		}
		printf ("FRAME %d, %d\n",images[i].frame_count,sizeof(images[i]));
	}
/*
	imagecount = writeBinary(images,imagecount,"images.let");
	image* newimages;
	imagecount = readBinary(&newimages);
*/


	spitOutPlain(images,imagecount,"images.let");
	/*readArrays( "images.let" );*/

	freeImageList(images,imagecount);

	return 0;
}
